# SPDX-License-Identifier: GPL-3.0+

"""
Tools for analyzing UEFI firmware using radare2
"""

import binascii
import json
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

import yaml

from fwhunt_scan.uefi_analyzer import (NvramVariable, UefiAnalyzer, UefiGuid,
                                       UefiProtocol, UefiService)

logger = logging.getLogger(__name__)


class CodePattern:
    """Code pattern"""

    def __init__(self, pattern: str, place: Optional[str]) -> None:
        self.pattern: str = str(pattern)
        self.place: Optional[str] = place
        # if True, scan the whole binary
        self.unspecified: bool = place is None

    @property
    def __dict__(self):
        return dict(
            {
                "pattern": self.pattern,
                "unspecified": self.unspecified,
            }
        )


class UefiRuleVariant:
    """A rule for scanning EFI image (content without meta and variants)"""

    def __init__(self, rule_content: Dict[str, Any]):
        self._uefi_rule: Dict[str, Any] = rule_content
        self._nvram_vars: Optional[Dict[str, List[NvramVariable]]] = None
        self._protocols: Optional[Dict[str, List[UefiProtocol]]] = None
        self._ppi_list: Optional[Dict[str, List[UefiProtocol]]] = None
        self._guids: Optional[Dict[str, List[UefiGuid]]] = None
        self._strings: Optional[Dict[str, List[str]]] = None
        self._wide_strings: Optional[Dict[str, List[Dict[str, str]]]] = None
        self._hex_strings: Optional[Dict[str, List[str]]] = None
        self._code: Optional[Dict[str, List[CodePattern]]] = None

    def _get_code(self) -> Dict[str, List[CodePattern]]:
        code: Dict[str, List[CodePattern]] = dict()
        if "code" not in self._uefi_rule:
            return code
        dict_items = dict()
        if isinstance(self._uefi_rule["code"], list):
            # if kind of matches is not specified
            dict_items["and"] = self._uefi_rule["code"]
        elif isinstance(self._uefi_rule["code"], dict):
            dict_items = self._uefi_rule["code"]
        else:
            return code
        for op in dict_items:
            code[op] = list()
            for c in dict_items[op]:
                cp = CodePattern(
                    pattern=c.get("pattern", None),
                    place=c.get("place", None),
                )
                code[op].append(cp)
        return code

    @property
    def code(self) -> Dict[str, List[CodePattern]]:
        """Get code from rule"""

        if self._code is None:
            self._code = self._get_code()
        return self._code

    def _get_strings(self) -> Dict[str, List[str]]:
        strings: Dict[str, List[str]] = dict()
        if "strings" not in self._uefi_rule:
            return strings
        if isinstance(self._uefi_rule["strings"], list):
            # if kind of matches is not specified
            strings["and"] = self._uefi_rule["strings"]
        elif isinstance(self._uefi_rule["strings"], dict):
            strings = self._uefi_rule["strings"]
        else:
            return strings
        return strings

    @property
    def strings(self) -> Dict[str, List[str]]:
        """Get strings from rule"""

        if self._strings is None:
            self._strings = self._get_strings()
        return self._strings

    def _get_wide_strings(self) -> Dict[str, List[Dict[str, str]]]:
        wide_strings: Dict[str, List[Dict[str, str]]] = dict()
        if "wide_strings" not in self._uefi_rule:
            return wide_strings
        if isinstance(self._uefi_rule["wide_strings"], list):
            # if kind of matches is not specified
            wide_strings["and"] = self._uefi_rule["wide_strings"]
        elif isinstance(self._uefi_rule["wide_strings"], dict):
            wide_strings = self._uefi_rule["wide_strings"]
        else:
            return wide_strings
        return wide_strings

    @property
    def wide_strings(self) -> Dict[str, List[Dict[str, str]]]:
        """Get wide strings from rule"""

        if self._wide_strings is None:
            self._wide_strings = self._get_wide_strings()
        return self._wide_strings

    def _get_hex_strings(self) -> Dict[str, List[str]]:
        hex_strings: Dict[str, List[str]] = dict()
        if "hex_strings" not in self._uefi_rule:
            return hex_strings
        if isinstance(self._uefi_rule["hex_strings"], list):
            # if kind of matches is not specified
            hex_strings["and"] = self._uefi_rule["hex_strings"]
        elif isinstance(self._uefi_rule["hex_strings"], dict):
            hex_strings = self._uefi_rule["hex_strings"]
        else:
            return hex_strings
        return hex_strings

    @property
    def hex_strings(self) -> Dict[str, List[str]]:
        """Get hex strings from rule"""

        if self._hex_strings is None:
            self._hex_strings = self._get_hex_strings()
        return self._hex_strings

    def _get_nvram_vars(self) -> Dict[str, List[NvramVariable]]:
        nvram_vars: Dict[str, List[NvramVariable]] = dict()
        if "nvram" not in self._uefi_rule:
            return nvram_vars
        dict_items = dict()
        if isinstance(self._uefi_rule["nvram"], list):
            # if kind of matches is not specified
            dict_items["and"] = self._uefi_rule["nvram"]
        elif isinstance(self._uefi_rule["nvram"], dict):
            dict_items = self._uefi_rule["nvram"]
        else:
            return nvram_vars
        for op in dict_items:
            nvram_vars[op] = list()
            for element in dict_items[op]:
                nvram_vars[op].append(
                    NvramVariable(
                        name=element["name"],
                        guid=element["guid"],
                        service=UefiService(
                            name=element["service"]["name"], address=0x0
                        ),
                    )
                )
        return nvram_vars

    @property
    def nvram_vars(self) -> Dict[str, List[NvramVariable]]:
        """Get NVRAM variables from rule"""

        if self._nvram_vars is None:
            self._nvram_vars = self._get_nvram_vars()
        return self._nvram_vars

    def _get_protocols(self) -> Dict[str, List[UefiProtocol]]:
        protocols: Dict[str, List[UefiProtocol]] = dict()
        if "protocols" not in self._uefi_rule:
            return protocols
        dict_items = dict()
        if isinstance(self._uefi_rule["protocols"], list):
            # if kind of matches is not specified
            dict_items["and"] = self._uefi_rule["protocols"]
        elif isinstance(self._uefi_rule["protocols"], dict):
            dict_items = self._uefi_rule["protocols"]
        else:
            return protocols
        for op in dict_items:
            protocols[op] = list()
            for element in dict_items[op]:
                protocols[op].append(
                    UefiProtocol(
                        name=element["name"],
                        value=element["value"],
                        service=element["service"]["name"],
                        address=0x0,
                        guid_address=0x0,
                    )
                )
        return protocols

    @property
    def protocols(self) -> Dict[str, List[UefiProtocol]]:
        """Get protocols from rule"""

        if self._protocols is None:
            self._protocols = self._get_protocols()
        return self._protocols

    def _get_ppi_list(self) -> Dict[str, List[UefiProtocol]]:
        ppi_list: Dict[str, List[UefiProtocol]] = dict()
        if "ppi" not in self._uefi_rule:
            return ppi_list
        dict_items = dict()
        if isinstance(self._uefi_rule["ppi"], list):
            # if kind of matches is not specified
            dict_items["and"] = self._uefi_rule["ppi"]
        elif isinstance(self._uefi_rule["ppi"], dict):
            dict_items = self._uefi_rule["ppi"]
        else:
            return ppi_list
        for op in dict_items:
            ppi_list[op] = list()
            for element in dict_items[op]:
                ppi_list[op].append(
                    UefiProtocol(
                        name=element["name"],
                        value=element["value"],
                        service=element["service"]["name"],
                        address=0x0,
                        guid_address=0x0,
                    )
                )
        return ppi_list

    @property
    def ppi_list(self) -> Dict[str, List[UefiProtocol]]:
        """Get PPI list from rule"""

        if self._ppi_list is None:
            self._ppi_list = self._get_ppi_list()
        return self._ppi_list

    def _get_guids(self) -> Dict[str, List[UefiGuid]]:
        guids: Dict[str, List[UefiGuid]] = dict()
        if "guids" not in self._uefi_rule:
            return guids
        dict_items = dict()
        if isinstance(self._uefi_rule["guids"], list):
            # if kind of matches is not specified
            dict_items["and"] = self._uefi_rule["guids"]
        elif isinstance(self._uefi_rule["guids"], dict):
            dict_items = self._uefi_rule["guids"]
        else:
            return guids
        for op in dict_items:
            guids[op] = list()
            for element in dict_items[op]:
                guids[op].append(
                    UefiGuid(
                        name=element["name"],
                        value=element["value"],
                    )
                )
        return guids

    @property
    def guids(self) -> Dict[str, List[UefiGuid]]:
        """Get GUIDs from rule"""

        if self._guids is None:
            self._guids = self._get_guids()
        return self._guids


class UefiRule:
    """A rule for scanning EFI image"""

    def __init__(
        self, rule_path: Optional[str] = None, rule_content: Optional[str] = None
    ):
        self._rule: Optional[str] = rule_path
        self._rule_name: str = str()
        self._uefi_rule: Dict[str, Any] = dict()
        if self._rule is not None:
            if os.path.isfile(self._rule):
                try:
                    with open(self._rule, "r") as f:
                        self._uefi_rule = yaml.safe_load(f)
                except yaml.scanner.ScannerError as e:
                    logger.error(repr(e))
        elif rule_content is not None:
            try:
                self._uefi_rule = yaml.safe_load(rule_content)
            except yaml.scanner.ScannerError as e:
                logger.error(repr(e))
        if self._uefi_rule:
            self._rule_name = list(self._uefi_rule.keys())[0]
        if self._rule_name:
            self._uefi_rule = self._uefi_rule[self._rule_name]
        self._variants: Optional[Dict[str, UefiRuleVariant]] = None

    def __str__(self):
        return json.dumps(self._uefi_rule, indent=2)

    @property
    def author(self) -> Optional[str]:
        """Get author from the metadata block"""

        try:
            return self._uefi_rule["meta"]["author"]
        except KeyError:
            return None

    @property
    def name(self) -> Optional[str]:
        """Get rule name from the metadata block"""

        try:
            return self._uefi_rule["meta"]["name"]
        except KeyError:
            return None

    @property
    def version(self) -> Optional[str]:
        """Get rule version from the metadata block"""

        try:
            return self._uefi_rule["meta"]["version"]
        except KeyError:
            return None

    @property
    def namespace(self) -> Optional[str]:
        """Get rule namespace from the metadata block"""

        try:
            return self._uefi_rule["meta"]["namespace"]
        except KeyError:
            return None

    @property
    def license(self) -> Optional[str]:
        """Get rule license from the metadata block"""

        try:
            return self._uefi_rule["meta"]["license"]
        except KeyError:
            return None

    @property
    def cve_number(self) -> Optional[str]:
        """Get CVE number from the metadata block"""

        try:
            return self._uefi_rule["meta"]["CVE number"]
        except KeyError:
            return None

    @property
    def vendor_id(self) -> Optional[str]:
        """Get vendor id from the metadata block"""

        try:
            return self._uefi_rule["meta"]["vendor id"]
        except KeyError:
            return None

    @property
    def cvss_score(self) -> Optional[str]:
        """Get CVSS score from the metadata block"""

        try:
            return self._uefi_rule["meta"]["CVSS score"]
        except KeyError:
            return None

    @property
    def advisory(self) -> Optional[str]:
        """Get advisory link from the metadata block"""

        try:
            return self._uefi_rule["meta"]["advisory"]
        except KeyError:
            return None

    @property
    def description(self) -> Optional[str]:
        """Get optional rule description from the metadata block"""

        try:
            return self._uefi_rule["meta"]["description"]
        except KeyError:
            return None

    @property
    def url(self) -> Optional[str]:
        """Get optional rule URL from the metadata block"""

        try:
            return self._uefi_rule["meta"]["url"]
        except KeyError:
            return None

    @property
    def target(self) -> Optional[str]:
        """Get optional rule target from the metadata block"""

        try:
            return self._uefi_rule["meta"]["target"]
        except KeyError:
            return None

    @property
    def volume_guids(self) -> List[str]:
        """Get any volume GUIDs from the metadata block"""

        try:
            return self._uefi_rule["meta"]["volume guids"]
        except KeyError:
            return list()

    def _get_variants(self) -> Dict[str, UefiRuleVariant]:
        """Get rules variants"""

        variants: Dict[str, UefiRuleVariant] = dict()
        if "variants" in self._uefi_rule:
            for variant in self._uefi_rule["variants"]:
                variants[variant] = UefiRuleVariant(
                    rule_content=self._uefi_rule["variants"][variant]
                )

        if "variants" not in self._uefi_rule:
            rule_content = {
                k: self._uefi_rule[k] for k in self._uefi_rule if k != "meta"
            }
            variants["default"] = UefiRuleVariant(rule_content=rule_content)

        return variants

    @property
    def variants(self) -> Dict[str, UefiRuleVariant]:
        """Get GUIDs from rule"""

        if self._variants is None:
            self._variants = self._get_variants()
        return self._variants


class UefiScannerError(Exception):
    """Generic scanner error exception."""

    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self):
        return repr(self.value)


class UefiScannerRes:
    """Scanner result for single rule"""

    def __init__(self, rule: UefiRule, variant_label: str, res: bool) -> None:
        self.rule = rule
        self.variant_label = variant_label
        self.res = res


class UefiScanner:
    """helper object for scanning an EFI image with multiple rules"""

    PROTOCOL: int = 0
    PPI: int = 1

    def __init__(self, uefi_analyzer: UefiAnalyzer, uefi_rules: List[UefiRule]):
        self._uefi_analyzer: UefiAnalyzer = uefi_analyzer
        self._uefi_rules: List[UefiRule] = uefi_rules
        self._valid_rules: bool = self._check_rules()
        if not self._valid_rules:
            raise UefiScannerError(
                "Invalid rule format. Visit https://github.com/binarly-io/FwHunt to find the latest version of the rules format."
            )
        self._results: Optional[List[UefiScannerRes]] = None

    def _check_rule(self, rule: UefiRule) -> bool:
        variants: Optional[Dict[str, UefiRuleVariant]] = None
        try:
            variants = rule.variants
        except KeyError:
            return False

        if not isinstance(variants, dict):
            return False

        for label in variants:
            if not isinstance(label, str):
                return False
            if not isinstance(variants[label], UefiRuleVariant):
                return False

        return True

    def _check_rules(self) -> bool:
        for rule in self._uefi_rules:
            if not self._check_rule(rule):
                return False
        return True

    def _and_strings(self, strings: List[str]) -> bool:
        res = True
        for string in strings:
            res &= not not self._uefi_analyzer._rz.cmdj("/j {}".format(string))
            if not res:
                break
        return res

    def _or_strings(self, strings: List[str]) -> bool:
        res = False
        for string in strings:
            res |= not not self._uefi_analyzer._rz.cmdj("/j {}".format(string))
            if res:
                break
        return res

    def _and_hex_strings(self, strings: List[str]) -> bool:
        res = True
        for string in strings:
            res &= not not self._uefi_analyzer._rz.cmdj("/xj {}".format(string))
            if not res:
                break
        return res

    def _or_hex_strings(self, strings: List[str]) -> bool:
        res = False
        for string in strings:
            res |= not not self._uefi_analyzer._rz.cmdj("/xj {}".format(string))
            if res:
                break
        return res

    def _and_wide_strings(self, strings: List[dict]) -> bool:
        res = True
        for item in strings:
            if "utf16le" in item:
                res &= not not self._uefi_analyzer._rz.cmdj(
                    "/wj {}".format(item["utf16le"])
                )
            elif "utf16be" in item:
                string = item["utf16be"]
                res &= not not self._uefi_analyzer._rz.cmdj(
                    "/xj {}".format(
                        binascii.hexlify(string.encode("utf-16be")).decode()
                    )
                )
            else:
                raise UefiScannerError("Wrong wide string format")

            if not res:
                break

        return res

    def _or_wide_strings(self, strings: List[dict]) -> bool:
        res = False
        for item in strings:
            if "utf16le" in item:
                res |= not not self._uefi_analyzer._rz.cmdj(
                    "/wj {}".format(item["utf16le"])
                )
            elif "utf16be" in item:
                string = item["utf16be"]
                res |= not not self._uefi_analyzer._rz.cmdj(
                    "/xj {}".format(
                        binascii.hexlify(string.encode("utf-16be")).decode()
                    )
                )
            else:
                raise UefiScannerError("Wrong wide string format")

            if res:
                break

        return res

    def _strings_scanner(self, rule_strings: Dict[str, List[str]]) -> bool:
        """Match strings"""

        final_res = True
        for op in rule_strings:
            # check kind of matches
            if op not in ["and", "or", "not-any", "not-all"]:
                raise UefiScannerError(
                    f"Invalid kind of matches: {op} (possible kinds of matches: and, or, not-any, not-all)"
                )

            res = True
            if op == "and":  # AND
                res = self._and_strings(rule_strings[op])

            if op == "or":  # OR
                res = self._or_strings(rule_strings[op])

            if op == "not-any":  # NOT OR
                res = not self._or_strings(rule_strings[op])

            if op == "not-all":  # NOT AND
                res = not self._and_strings(rule_strings[op])

            final_res &= res  # AND between all sets of strings
            if not final_res:
                break

        return final_res

    def _wide_strings_scanner(
        self, rule_wide_strings: Dict[str, List[Dict[str, str]]]
    ) -> bool:
        """Match wide strings"""

        final_res = True
        for op in rule_wide_strings:
            # check kind of matches
            if op not in ["and", "or", "not-any", "not-all"]:
                raise UefiScannerError(
                    f"Invalid kind of matches: {op} (possible kinds of matches: and, or, not-any, not-all)"
                )

            res = True
            if op == "and":  # AND
                res = self._and_wide_strings(rule_wide_strings[op])

            if op == "or":  # OR
                res = self._or_wide_strings(rule_wide_strings[op])

            if op == "not-any":  # NOT OR
                res = not self._or_wide_strings(rule_wide_strings[op])

            if op == "not-all":  # NOT AND
                res = not self._and_wide_strings(rule_wide_strings[op])

            final_res &= res  # AND between all sets of strings
            if not final_res:
                break

        return final_res

    def _hex_strings_scanner(self, rule_hex_strings: Dict[str, List[str]]) -> bool:
        """Match hex strings"""

        final_res = True
        for op in rule_hex_strings:
            # check kind of matches
            if op not in ["and", "or", "not-any", "not-all"]:
                raise UefiScannerError(
                    f"Invalid kind of matches: {op} (possible kinds of matches: and, or, not-any, not-all)"
                )

            res = True
            if op == "and":  # AND
                res = self._and_hex_strings(rule_hex_strings[op])

            if op == "or":  # OR
                res = self._or_hex_strings(rule_hex_strings[op])

            if op == "not-any":  # NOT OR
                res = not self._or_hex_strings(rule_hex_strings[op])

            if op == "not-all":  # NOT AND
                res = not self._and_hex_strings(rule_hex_strings[op])

            final_res &= res  # AND between all sets of strings
            if not final_res:
                break

        return final_res

    def _search_nvram(self, nvram_rule: NvramVariable) -> bool:
        for nvram_analyzer in self._uefi_analyzer.nvram_vars:
            if (
                nvram_rule.name == nvram_analyzer.name
                and nvram_rule.guid == nvram_analyzer.guid
                and nvram_rule.service.name == nvram_analyzer.service.name
            ):
                return True
        return False

    def _and_nvram(self, nvram_vars: List[NvramVariable]) -> bool:
        res = True
        for nvram_var in nvram_vars:
            res &= self._search_nvram(nvram_var)
            if not res:
                break
        return res

    def _or_nvram(self, nvram_vars: List[NvramVariable]) -> bool:
        res = False
        for nvram_var in nvram_vars:
            res |= self._search_nvram(nvram_var)
            if res:
                break
        return res

    def _compare_nvram_vars(self, rule_nvram: Dict[str, List[NvramVariable]]) -> bool:
        """Compare NVRAM variables"""

        final_res = True
        for op in rule_nvram:
            # check kind of matches
            if op not in ["and", "or", "not-any", "not-all"]:
                raise UefiScannerError(
                    f"Invalid kind of matches: {op} (possible kinds of matches: and, or, not-any, not-all)"
                )

            res = True
            if op == "and":  # AND
                res = self._and_nvram(rule_nvram[op])

            if op == "or":  # OR
                res = self._or_nvram(rule_nvram[op])

            if op == "not-any":  # NOT OR
                res = not self._or_nvram(rule_nvram[op])

            if op == "not-all":  # NOT AND
                res = not self._and_nvram(rule_nvram[op])

            final_res &= res  # AND between all sets of NVRAM variables
            if not final_res:
                break

        return final_res

    def _search_protocol(self, protocol_rule: UefiProtocol, mode: int) -> bool:
        items: List[UefiProtocol] = self._uefi_analyzer.protocols
        if mode == UefiScanner.PPI:
            items = self._uefi_analyzer.ppi_list
        for protocol_analyzer in items:
            if (
                protocol_rule.name == protocol_analyzer.name
                and protocol_rule.value == protocol_analyzer.value
                and protocol_rule.service == protocol_analyzer.service
            ):
                return True
        return False

    def _and_protocols(self, protocols: List[UefiProtocol], mode: int) -> bool:
        res = True
        for protocol in protocols:
            res &= self._search_protocol(protocol, mode)
            if not res:
                break
        return res

    def _or_protocols(self, protocols: List[UefiProtocol], mode: int) -> bool:
        res = False
        for protocol in protocols:
            res |= self._search_protocol(protocol, mode)
            if res:
                break
        return res

    def _compare_protocols(
        self, rule_protocol: Dict[str, List[UefiProtocol]], mode: int
    ) -> bool:
        """Compare protocols"""

        final_res = True
        for op in rule_protocol:
            # check kind of matches
            if op not in ["and", "or", "not-any", "not-all"]:
                raise UefiScannerError(
                    f"Invalid kind of matches: {op} (possible kinds of matches: and, or, not-any, not-all)"
                )

            res = True
            if op == "and":  # AND
                res = self._and_protocols(rule_protocol[op], mode)

            if op == "or":  # OR
                res = self._or_protocols(rule_protocol[op], mode)

            if op == "not-any":  # NOT OR
                res = not self._or_protocols(rule_protocol[op], mode)

            if op == "not-all":  # NOT AND
                res = not self._and_protocols(rule_protocol[op], mode)

            final_res &= res  # AND between all sets of protocols
            if not final_res:
                break

        return final_res

    def _search_guid(self, guid_rule: UefiGuid) -> bool:
        for guid in self._uefi_analyzer.protocol_guids:
            if guid_rule.name == guid.name or guid_rule.value == guid.value:
                # True if name or value is matches
                # so that UNKNOWN_GUIDs can be specified
                return True
        return False

    def _and_guids(self, guids: List[UefiGuid]) -> bool:
        res = True
        for guid in guids:
            res &= self._search_guid(guid)
            if not res:
                break
        return res

    def _or_guids(self, guids: List[UefiGuid]) -> bool:
        res = False
        for guid in guids:
            res |= self._search_guid(guid)
            if res:
                break
        return res

    def _compare_guids(self, rule_guid: Dict[str, List[UefiGuid]]) -> bool:
        """Compare GUIDs"""

        final_res = True
        for op in rule_guid:
            # check kind of matches
            if op not in ["and", "or", "not-any", "not-all"]:
                raise UefiScannerError(
                    f"Invalid kind of matches: {op} (possible kinds of matches: and, or, not-any, not-all)"
                )

            res = True
            if op == "and":  # AND
                res = self._and_guids(rule_guid[op])

            if op == "or":  # OR
                res = self._or_guids(rule_guid[op])

            if op == "not-any":  # NOT OR
                res = not self._or_guids(rule_guid[op])

            if op == "not-all":  # NOT AND
                res = not self._and_guids(rule_guid[op])

            final_res &= res  # AND between all sets of GUIDs
            if not final_res:
                break

        return final_res

    def _get_bounds(self, insns: List[Dict[str, Any]]) -> Tuple:
        """Get function end address"""

        funcs = list(
            filter(
                lambda addr: addr,
                [addr.get("offset", None) for addr in self._uefi_analyzer.functions],
            )
        )
        funcs.sort()

        start = insns[0].get("offset", None)
        end_insn = insns[-1].get("offset", None)

        if start is None or end_insn is None:
            return tuple((None, None))

        if start == funcs[-1]:
            return tuple((start, end_insn))

        try:
            start_index = funcs.index(start)
        except ValueError:
            return tuple((start, end_insn))

        end_func = funcs[start_index + 1]
        if end_insn < end_func:
            return tuple((start, end_insn))

        return tuple((start, end_func))

    @staticmethod
    def _tree_debug(start: int, end: int, depth: int) -> None:
        if not depth:
            logger.debug(
                f"\nFunction tree in the handler at {start:#x} (from {start:#x} to {end:#x})"
            )
        else:
            prefix = depth * "--"
            logger.debug(f"{prefix}{start:#x} (from {start:#x} to {end:#x})")

    def _get_bounds_rec(self, start_addr: int, depth: int, debug: bool) -> bool:
        """Recursively traverse the function and find the boundaries of all child functions"""

        self._uefi_analyzer._rz.cmd(f"s {start_addr:#x}")
        self._uefi_analyzer._rz.cmd("af")

        insns = self._uefi_analyzer._rz.cmd("pdrj")
        # prevent error messages to sys.stderr from rizin:
        # https://github.com/rizinorg/rz-pipe/blob/0f7ac66e6d679ebb03be26bf61a33f9ccf199f27/python/rzpipe/open_base.py#L261
        try:
            insns = json.loads(insns)
        except (ValueError, KeyError, TypeError):
            return False

        # append function bounds
        start, end = self._get_bounds(insns)

        if start is not None and end is not None and start_addr == start:
            self._funcs_bounds.append((start, end))

        if debug:
            self._tree_debug(start_addr, end, depth)
            depth += 1

        # scan child functions
        for insn in insns:
            if insn.get("type", None) != "call":
                continue
            if "esil" not in insn:
                continue
            esil = insn["esil"].split(",")
            if esil[-3:] != ["=[]", "rip", "="]:
                continue
            try:
                address = int(esil[0])
                if address not in self._rec_addrs:
                    self._rec_addrs.append(address)
                    self._get_bounds_rec(address, depth, debug)
            except ValueError:
                continue

        return True

    def _hex_strings_scanner_bounds(self, pattern: str, start: int, end: int) -> bool:
        """Match hex strings"""

        res = self._uefi_analyzer._rz.cmdj(f"/xj {pattern}")
        if not res:
            return False

        for sres in res:
            offset = sres.get("offset", None)
            if offset is None:
                continue

            if offset >= start and offset <= end:
                return True

        return False

    def _clear_cache(self) -> None:
        self._funcs_bounds: List[Any] = list()
        self._rec_addrs: List[Any] = list()

    def _code_scan_rec(self, address: int, pattern: str) -> bool:
        self._clear_cache()

        self._get_bounds_rec(address, depth=0, debug=False)
        if not len(self._funcs_bounds):
            return False

        for start, end in self._funcs_bounds:
            if self._hex_strings_scanner_bounds(pattern, start, end):
                return True

        return False

    def _get_handlers(self, c: CodePattern) -> List[Any]:
        if c.place == "child_sw_smi_handlers":
            return self._uefi_analyzer.child_swsmi_handlers
        if c.place == "smi_handlers":
            return self._uefi_analyzer.smi_handlers
        return [
            handler
            for handler in self._uefi_analyzer.smi_handlers
            if handler.place == c.place
        ]

    def _single_code_scan(self, c: CodePattern) -> bool:
        search_res = False
        if c.unspecified:
            return not not self._uefi_analyzer._rz.cmdj("/xj {}".format(c.pattern))
        for handler in self._get_handlers(c):
            search_res = self._code_scan_rec(handler.address, c.pattern)
            if search_res:
                break
        return search_res

    def _and_code(self, cs: List[CodePattern]) -> bool:
        res = True
        for c in cs:
            res &= self._single_code_scan(c)
            if not res:
                break

        return res

    def _or_code(self, cs: List[CodePattern]) -> bool:
        res = False
        for c in cs:
            res |= self._single_code_scan(c)
            if res:
                break

        return res

    def _code_scanner(self, rule_code: Dict[str, List[CodePattern]]) -> bool:
        """Compare code patterns"""

        final_res = True
        for op in rule_code:
            # check kind of matches
            if op not in ["and", "or", "not-any", "not-all"]:
                raise UefiScannerError(
                    f"Invalid kind of matches: {op} (possible kinds of matches: and, or, not-any, not-all)"
                )

            res = True
            if op == "and":  # AND
                res = self._and_code(rule_code[op])

            if op == "or":  # OR
                res = self._or_code(rule_code[op])

            if op == "not-any":  # NOT OR
                res = not self._or_code(rule_code[op])

            if op == "not-all":  # NOT AND
                res = not self._and_code(rule_code[op])

            final_res &= res  # AND between all sets of NVRAM variables
            if not final_res:
                break

        return final_res

    def _get_results_variants(
        self, rule_variant: UefiRuleVariant, target: Optional[str]
    ) -> bool:
        if target in ["firmware"]:
            # match hex strings
            return self._hex_strings_scanner(rule_variant.hex_strings)

        res = True

        # compare NVRAM variables
        res &= self._compare_nvram_vars(rule_variant.nvram_vars)
        if not res:
            return res

        # compare protocols
        res &= self._compare_protocols(rule_variant.protocols, UefiScanner.PROTOCOL)
        if not res:
            return res

        # compare GUIDs
        res &= self._compare_guids(rule_variant.guids)
        if not res:
            return res

        # compare PPI
        res &= self._compare_protocols(rule_variant.ppi_list, UefiScanner.PPI)
        if not res:
            return res

        # match code patterns
        res &= self._code_scanner(rule_variant.code)
        if not res:
            return res

        # match strings
        res &= self._strings_scanner(rule_variant.strings)
        if not res:
            return res

        # match wide strings
        res &= self._wide_strings_scanner(rule_variant.wide_strings)
        if not res:
            return res

        # match hex strings
        res &= self._hex_strings_scanner(rule_variant.hex_strings)
        if not res:
            return res

        return res

    def _get_results(self) -> List[UefiScannerRes]:
        results: List[UefiScannerRes] = list()

        for uefi_rule in self._uefi_rules:
            for variant in uefi_rule.variants:
                res = self._get_results_variants(
                    uefi_rule.variants[variant], uefi_rule.target
                )
                results.append(
                    UefiScannerRes(rule=uefi_rule, variant_label=variant, res=res)
                )

        return results

    @property
    def results(self) -> List[UefiScannerRes]:
        """Get scanning results as a list of matched rules"""

        if self._results is None:
            self._results = self._get_results()

        return self._results



# SPDX-License-Identifier: GPL-3.0+

"""
Tools for analyzing UEFI firmware and checking UEFI modules with FwHunt rules
"""

from .uefi_analyzer import UefiAnalyzer, UefiAnalyzerError
from .uefi_extractor import UefiBinary, UefiExtractor
from .uefi_scanner import UefiRule, UefiScanner, UefiScannerError
from .uefi_te import TerseExecutableParser

__all__ = [
    "UefiAnalyzer",
    "UefiRule",
    "UefiScanner",
    "UefiScannerError",
    "TerseExecutableParser",
    "UefiAnalyzerError",
    "UefiBinary",
    "UefiExtractor",
]



#!/usr/bin/env python3
#
# SPDX-License-Identifier: GPL-3.0+
#
# pylint: disable=invalid-name,missing-module-docstring,missing-function-docstring

# fwhunt_scan: tools for analyzing UEFI firmware and checking UEFI modules with FwHunt rules

import json
import logging
import os
import pathlib
import tempfile
from typing import Dict, List

import click

from fwhunt_scan import UefiAnalyzer, UefiExtractor, UefiRule, UefiScanner

logging.basicConfig(
    format="%(name)s %(asctime)s %(levelname)s: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
)
logger = logging.getLogger("fwhunt_scan")


@click.group()
def cli():
    pass


@click.command()
@click.argument("path")
@click.option("-o", "--out", help="Output JSON file.")
def analyze(path: str, out: str) -> bool:
    """Analyze single EFI file."""

    if not os.path.isfile(path):
        click.echo(
            "{} check module path".format(click.style("ERROR", fg="red", bold=True))
        )
        return False

    # on linux platforms you can pass blob via shm://
    # uefi_analyzer = UefiAnalyzer(blob=data)

    summary = None
    with UefiAnalyzer(image_path=path) as uefi_analyzer:
        summary = uefi_analyzer.get_summary()

    if out:
        with open(out, "w") as f:
            json.dump(summary, f, indent=4)
    else:
        click.echo(json.dumps(summary, indent=4))

    return True


@click.command()
@click.argument("path")
@click.option("-r", "--rule", help="The path to the rule.", multiple=True)
def scan(path: str, rule: List[str]) -> bool:
    """Scan single EFI file."""

    rules = rule

    if not os.path.isfile(path):
        click.echo(
            "{} check module path".format(click.style("ERROR", fg="red", bold=True))
        )
        return False
    if not all(rule and os.path.isfile(rule) for rule in rules):
        click.echo(
            "{} check rule(s) path".format(click.style("ERROR", fg="red", bold=True))
        )
        return False

    # on linux platforms you can pass blob via shm://
    # uefi_analyzer = UefiAnalyzer(blob=data)

    uefi_analyzer = UefiAnalyzer(image_path=path)

    uefi_rules: List[UefiRule] = list()

    for r in rules:
        with open(r, "r") as f:
            rule_content = f.read()
            uefi_rule = UefiRule(rule_content=rule_content)
            uefi_rules.append(uefi_rule)

    scanner = UefiScanner(uefi_analyzer, uefi_rules)
    prefix = click.style("Scanner result", fg="green")

    no_threat = click.style("No threat detected", fg="green")
    threat = click.style(
        "FwHunt rule has been triggered and threat detected!", fg="red"
    )

    for result in scanner.results:
        msg = threat if result.res else no_threat
        click.echo(
            f"{prefix} {result.rule.name} (variant: {result.variant_label}) {msg} ({path})"
        )

    uefi_analyzer.close()

    return True


@click.command()
@click.argument("image_path")
@click.option("-r", "--rule", help="The path to the rule.", multiple=True)
@click.option("-d", "--rules_dir", help="The path to the rules directory.")
@click.option(
    "-f",
    "--force",
    help="Enforcing the use of rules without specified volume guids.",
    is_flag=True,
)
def scan_firmware(
    image_path: str, rule: List[str], rules_dir: str, force: bool
) -> bool:
    """Scan UEFI firmware image."""

    rules = list(rule)
    error_prefix = click.style("ERROR", fg="red", bold=True)
    if not rules_dir:
        if not all(rules and os.path.isfile(rule) for rule in rules):
            click.echo(f"{error_prefix} check rule(s) path")
            return False
    else:
        rules += list(map(str, pathlib.Path(rules_dir).rglob("*.yml")))

    if not os.path.isfile(image_path):
        click.echo(f"{error_prefix} check image path")
        return False

    prefix = click.style("Scanner result", fg="green")
    no_threat = click.style("No threat detected", fg="green")
    threat = click.style(
        "FwHunt rule has been triggered and threat detected!", fg="red"
    )

    # on linux platforms you can pass blob via shm://
    # uefi_analyzer = UefiAnalyzer(blob=data)

    uefi_rules: List[UefiRule] = list()

    for r in rules:
        with open(r, "r") as f:
            uefi_rules.append(UefiRule(rule_content=f.read()))

    # Select the rules with `target: firmware`
    uefi_rules_fw: List[UefiRule] = list(
        filter(lambda rule: rule.target == "firmware", uefi_rules)
    )
    if len(uefi_rules_fw):
        with UefiAnalyzer(image_path=image_path) as uefi_analyzer_fw:
            scanner_fw = UefiScanner(uefi_analyzer_fw, uefi_rules_fw)
            for result in scanner_fw.results:
                msg = threat if result.res else no_threat
                click.echo(
                    f"{prefix} {result.rule.name} (variant: {result.variant_label}) {msg}"
                )

    # Group rules by guids
    rules_guids: Dict[str, List[UefiRule]] = dict()
    rules_universal: List[UefiRule] = list()
    for uefi_rule in set(uefi_rules) - set(uefi_rules_fw):
        if uefi_rule.target not in (None, "module"):
            logger.debug(
                f"The rule {uefi_rule.name} incompatible with scan-firmware command (target: {uefi_rule.target})"
            )
            continue
        if not len(uefi_rule.volume_guids) and not force:
            logger.warning(
                f"Specify volume_guids in {uefi_rule.name} or run command with --force flag"
            )
            continue
        elif not len(uefi_rule.volume_guids):
            rules_universal.append(uefi_rule)
        for guid in [g.lower() for g in uefi_rule.volume_guids]:
            lower_guid = guid.lower()
            if lower_guid not in rules_guids:
                rules_guids[lower_guid] = list()
            rules_guids[lower_guid].append(uefi_rule)

    if not rules_guids.keys() and not force:
        click.echo(
            f"{error_prefix} None of the rules specify volume_guids (use scan-module command)"
        )
        return False

    with open(image_path, "rb") as f:
        firmware_data = f.read()

    extractor = UefiExtractor(
        firmware_data, list(rules_guids.keys()) if not force else list()
    )
    extractor.extract_all(ignore_guid=force)

    if not len(extractor.binaries):
        click.echo("No modules were found for scanning")
        return False

    for binary in extractor.binaries:
        if not binary.is_ok:
            continue

        rules_scan = rules_universal + (
            rules_guids[binary.guid] if binary.guid in rules_guids else list()
        )

        if not len(rules_scan):
            continue

        fpath = os.path.join(tempfile.gettempdir(), f"{binary.name}{binary.ext}")
        with open(fpath, "wb") as f:
            f.write(binary.content)

        logger.debug(f"Scanning the module {binary.name}{binary.ext}")

        with UefiAnalyzer(image_path=fpath) as uefi_analyzer:
            scanner = UefiScanner(uefi_analyzer, rules_scan)
            for result in scanner.results:
                msg = threat if result.res else no_threat
                click.echo(
                    f"{prefix} {result.rule.name} (variant: {result.variant_label}) {msg} ({binary.name})"
                )

        os.remove(fpath)

    return True


@click.command()
@click.argument(
    "image_path", type=click.Path(exists=True, dir_okay=False, file_okay=True)
)
@click.argument("extract_path", type=click.Path(dir_okay=True, file_okay=False))
def extract(image_path: str, extract_path: str) -> bool:
    """Extract all modules from UEFI firmware image."""

    if not os.path.isdir(extract_path):
        os.mkdir(extract_path)

    with open(image_path, "rb") as f:
        firmware_data = f.read()

    extractor = UefiExtractor(firmware_data, list())
    extractor.extract_all(ignore_guid=True)

    if not len(extractor.binaries):
        click.echo("No modules found", err=True)
        return False

    for binary in extractor.binaries:
        if not binary.is_ok:
            continue
        fpath = os.path.join(extract_path, f"{binary.name}-{binary.guid}{binary.ext}")
        with open(fpath, "wb") as f:
            f.write(binary.content)

        click.echo(f"{binary.guid}: {fpath}")

    return True


cli.add_command(analyze)
cli.add_command(analyze, "analyze-module")
cli.add_command(analyze, "analyze-bootloader")
cli.add_command(scan)
cli.add_command(scan, "scan-module")
cli.add_command(scan, "scan-bootloader")
cli.add_command(scan_firmware)
cli.add_command(extract)

if __name__ == "__main__":
    cli()



# SPDX-License-Identifier: GPL-3.0+

import os
import struct
from typing import Optional


class TerseExecutableError(Exception):
    """Generic TE format error exception."""

    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self):
        return repr(self.value)


class TerseExecutableParser:
    """Terse Executable header parser"""

    def __init__(
        self, image_path: Optional[str] = None, blob: Optional[bytes] = None
    ) -> None:

        self._data: bytes = bytes()

        if blob is not None:
            self._data = blob

        elif image_path is not None:

            # check file path
            if not os.path.join(image_path):
                raise TerseExecutableError("Wrong file path")

            self._data = bytes()
            with open(image_path, "rb") as f:
                self._data = f.read()

        self._signature: bytes = self._data[:2]
        if self._signature != b"VZ":
            raise TerseExecutableError("Wrong signature")

        self._parse()

    def _parse(self) -> None:
        format = "<2sHBBHIIQ"
        data = self._data[: struct.calcsize(format)]
        if len(data) != struct.calcsize(format):
            raise TerseExecutableError("Can't parse header, file is too small")
        (
            self._signature,
            self._machine,
            self._number_of_sections,
            self._subsystem,
            self._stripped_size,
            self._address_of_entry_point,
            self._base_of_code,
            self._image_base,
        ) = struct.unpack(format, data)

    @property
    def signature(self) -> str:
        """Get Signature"""

        return self._signature.decode()

    @property
    def machine(self) -> int:
        """Get Machine"""

        return self._machine

    @property
    def number_of_sections(self) -> int:
        """Get NumberOfSections"""

        return self._number_of_sections

    @property
    def subsystem(self) -> int:
        """Get Subsystem"""

        return self._subsystem

    @property
    def stripped_size(self) -> int:
        """Get StrippedSize"""

        return self._stripped_size

    @property
    def address_of_entry_point(self) -> int:
        """Get AddressOfEntryPoint"""

        return self._address_of_entry_point

    @property
    def base_of_code(self) -> int:
        """Get BaseOfCode"""

        return self._base_of_code

    @property
    def image_base(self) -> int:
        """Get ImageBase"""

        return self._image_base

    def __str__(self):
        return "\n".join(
            [
                f"Machine: {self.machine:#x}",
                f"NumberOfSections = {self.number_of_sections:#x}",
                f"Subsystem = {self.subsystem:#x}",
                f"StrippedSize = {self.stripped_size:#x}",
                f"AddressOfEntryPoint = {self.address_of_entry_point:#x}",
                f"BaseOfCode = {self.base_of_code:#x}",
                f"ImageBase = {self.image_base:#x}",
            ]
        )



from types import TracebackType
from typing import Any, Dict, List, Optional, Type

from _typeshed import Incomplete

from fwhunt_scan.uefi_protocols import GUID_FROM_BYTES as GUID_FROM_BYTES
from fwhunt_scan.uefi_tables import \
    BS_PROTOCOLS_INFO_64_BIT as BS_PROTOCOLS_INFO_64_BIT
from fwhunt_scan.uefi_tables import \
    EFI_BOOT_SERVICES_64_BIT as EFI_BOOT_SERVICES_64_BIT
from fwhunt_scan.uefi_tables import \
    EFI_PEI_SERVICES_32_BIT as EFI_PEI_SERVICES_32_BIT
from fwhunt_scan.uefi_tables import \
    EFI_RUNTIME_SERVICES_64_BIT as EFI_RUNTIME_SERVICES_64_BIT
from fwhunt_scan.uefi_tables import OFFSET_TO_SERVICE as OFFSET_TO_SERVICE
from fwhunt_scan.uefi_te import TerseExecutableError as TerseExecutableError
from fwhunt_scan.uefi_te import TerseExecutableParser as TerseExecutableParser
from fwhunt_scan.uefi_types import ChildSwSmiHandler as ChildSwSmiHandler
from fwhunt_scan.uefi_types import NvramVariable as NvramVariable
from fwhunt_scan.uefi_types import SmiHandler as SmiHandler
from fwhunt_scan.uefi_types import UefiGuid as UefiGuid
from fwhunt_scan.uefi_types import UefiProtocol as UefiProtocol
from fwhunt_scan.uefi_types import UefiProtocolGuid as UefiProtocolGuid
from fwhunt_scan.uefi_types import UefiService as UefiService
from fwhunt_scan.uefi_utils import \
    get_current_insn_index as get_current_insn_index
from fwhunt_scan.uefi_utils import get_int as get_int

class UefiAnalyzerError(Exception):
    value: Incomplete
    def __init__(self, value: str) -> None: ...

class UefiAnalyzer:
    def __init__(self, image_path: Optional[str] = ..., blob: Optional[bytes] = ..., rizinhome: Optional[str] = ...) -> None: ...
    def __enter__(self): ...
    @property
    def info(self) -> Dict[Any, Any]: ...
    @property
    def strings(self) -> List[Any]: ...
    @property
    def sections(self) -> List[Any]: ...
    @property
    def functions(self) -> List[Any]: ...
    @property
    def insns(self) -> List[Any]: ...
    @property
    def g_bs(self) -> List[int]: ...
    @property
    def g_rt(self) -> List[int]: ...
    @property
    def boot_services(self) -> List[UefiService]: ...
    @property
    def boot_services_protocols(self) -> List[Any]: ...
    @property
    def runtime_services(self) -> List[UefiService]: ...
    @property
    def protocols(self) -> List[UefiProtocol]: ...
    @property
    def protocol_guids(self) -> List[UefiProtocolGuid]: ...
    def r2_get_nvram_vars_64bit(self) -> List[NvramVariable]: ...
    @property
    def nvram_vars(self) -> List[NvramVariable]: ...
    @property
    def pei_services(self) -> List[UefiService]: ...
    @property
    def ppi_list(self) -> List[UefiProtocol]: ...
    @property
    def smi_handlers(self) -> List[SmiHandler]: ...
    @property
    def child_swsmi_handlers(self) -> List[ChildSwSmiHandler]: ...
    @property
    def smst_list(self) -> List[int]: ...
    def get_summary(self) -> Dict[str, Any]: ...
    def get_protocols_info(self) -> Dict[str, Any]: ...
    def close(self) -> None: ...
    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], traceback: Optional[TracebackType]) -> None: ...



from typing import Any, Dict, List, Optional

import rzpipe

from fwhunt_scan.uefi_protocols import UefiGuid as UefiGuid

def get_int(item: str) -> Optional[int]: ...
def get_xrefs_to_guids(rz: rzpipe.open, guids: List[UefiGuid]) -> List[int]: ...
def get_xrefs_to_data(rz: rzpipe.open, addr: int) -> List[int]: ...
def get_current_insn_index(insns: Optional[List[Dict[str, Any]]], code_addr: int) -> Optional[int]: ...



from .uefi_analyzer import UefiAnalyzer as UefiAnalyzer
from .uefi_analyzer import UefiAnalyzerError as UefiAnalyzerError
from .uefi_extractor import UefiBinary as UefiBinary
from .uefi_extractor import UefiExtractor as UefiExtractor
from .uefi_scanner import UefiRule as UefiRule
from .uefi_scanner import UefiScanner as UefiScanner
from .uefi_scanner import UefiScannerError as UefiScannerError
from .uefi_te import TerseExecutableParser as TerseExecutableParser



from typing import List, Optional

from _typeshed import Incomplete

class UefiBinary:
    KINDS: Incomplete
    guid: Incomplete
    def __init__(self, content: Optional[bytes], name: Optional[str], guid: str, ext: Optional[str]) -> None: ...
    @property
    def content(self) -> bytes: ...
    @property
    def name(self) -> str: ...
    @property
    def ext(self) -> str: ...
    @property
    def is_ok(self) -> bool: ...

class UefiExtractor:
    FILE_TYPES: Incomplete
    SECTION_TYPES: Incomplete
    UI: Incomplete
    binaries: Incomplete
    def __init__(self, firmware_data: bytes, file_guids: List[str]) -> None: ...
    def extract_all(self, ignore_guid: bool = ...) -> None: ...



from typing import Dict

from _typeshed import Incomplete

from fwhunt_scan.uefi_types import UefiGuid as UefiGuid

PROTOCOLS_GUIDS: Incomplete
GUID_FROM_VALUE: Dict[str, UefiGuid]
GUID_FROM_BYTES: Dict[bytes, UefiGuid]



import contextlib
import os
from typing import Any, Dict, List, Optional

import uefi_firmware


class UefiBinary:
    KINDS = [
        ".sec",
        ".pei.core",
        ".dxe.core",
        ".peim",
        ".dxe",
        ".peim.dxe",
        ".app",
        ".smm",
        ".smm.dxe",
        ".smm.core",
    ]

    def __init__(
        self,
        content: Optional[bytes],
        name: Optional[str],
        guid: str,
        ext: Optional[str],
    ) -> None:
        self.guid: str = guid.lower()
        self._content: Optional[bytes] = content
        self._name: Optional[str] = name
        self._ext: Optional[str] = ext

    @property
    def content(self) -> bytes:
        if self._content is None:
            self._content = bytes()
        return self._content

    @property
    def name(self) -> str:
        if self._name is None:
            self._name = self.guid
        return self._name

    @property
    def ext(self) -> str:
        if self._ext is None:
            self._ext = ".bin"
        return self._ext

    @property
    def is_ok(self) -> bool:
        return self.guid and len(self.content) and self.ext in UefiBinary.KINDS


class UefiExtractor:
    FILE_TYPES = {
        0x01: ("raw", "raw", "RAW"),
        0x02: ("freeform", "freeform", "FREEFORM"),
        0x03: ("security core", "sec", "SEC"),
        0x04: ("pei core", "pei.core", "PEI_CORE"),
        0x05: ("dxe core", "dxe.core", "DXE_CORE"),
        0x06: ("pei module", "peim", "PEIM"),
        0x07: ("driver", "dxe", "DRIVER"),
        0x08: ("combined pei module/driver", "peim.dxe", "COMBO_PEIM_DRIVER"),
        0x09: ("application", "app", "APPLICATION"),
        0x0A: ("system management", "smm", "SMM"),
        0x0C: ("combined smm/driver", "smm.dxe", "COMBO_SMM_DRIVER"),
        0x0D: ("smm core", "smm.core", "SMM_CORE"),
    }
    SECTION_TYPES = {
        0x10: ("PE32 image", "pe", "PE32"),
        0x12: ("Terse executable (TE)", "te", "TE"),
    }
    UI = {0x15: ("User interface name", "ui", "UI")}

    def __init__(self, firmware_data: bytes, file_guids: List[str]):
        offset = firmware_data.find(b"_FVH")
        if offset >= 0:
            firmware_data = firmware_data[offset - 40 :]
        self._firmware_data: bytes = firmware_data
        self._file_guids: List[str] = [g.lower() for g in file_guids]
        self._parsers: List[uefi_firmware.AutoParser] = list()
        self._info: Dict[str, Any] = dict()
        self.binaries: List[UefiBinary] = list()

    def _compressed_search(self, object: Any, root_guid: str) -> None:
        if object is None:
            return

        for component in object.iterate_objects():
            attrs = component.get("attrs", None)
            if attrs is not None:
                section_type = attrs.get("type", None)
                if section_type in UefiExtractor.UI:
                    self._info[root_guid]["name"] = component["label"]
                if section_type in UefiExtractor.SECTION_TYPES:
                    self._info[root_guid]["content"] = component["_self"].content
            self._compressed_search(component["_self"], root_guid)

    def _compressed_handle(self, object: Any, root_guid: str) -> None:
        if object is None:
            return

        for obj in object.iterate_objects():
            if (
                obj.get("attrs", None) is not None
                and obj["attrs"].get("attrs", None) == 0x01
            ):  # if compressed
                self._compressed_search(obj["_self"], root_guid)

    def _append_binaries(self, object: Any) -> None:
        if object is None:
            return

        for component in object.iterate_objects():
            guid = component.get("guid", None)
            attrs = component.get("attrs", None)
            if guid is not None and attrs is not None:
                if guid not in self._info:
                    self._info[guid] = {"name": None, "ext": None, "content": None}
                section_type = attrs.get("type", None)
                if section_type in UefiExtractor.UI:
                    self._info[guid]["name"] = component["label"]
                if section_type in UefiExtractor.FILE_TYPES:
                    if self._info[guid]["ext"] is None:
                        ext = UefiExtractor.FILE_TYPES[section_type][1]
                        self._info[guid]["ext"] = f".{ext}"
                    self._compressed_handle(component["_self"], guid)
                if section_type in UefiExtractor.SECTION_TYPES:
                    self._info[guid]["content"] = component["_self"].content
            self._append_binaries(component["_self"])

    def _extract(self) -> bool:
        potencial_volumes = uefi_firmware.search_firmware_volumes(self._firmware_data)
        for offset in potencial_volumes:
            parser = uefi_firmware.AutoParser(self._firmware_data[offset - 40 :])
            if parser is None or parser.type() == "unknown":
                continue
            self._parsers.append(parser)

        if not len(self._parsers):
            return False

        for parser in self._parsers:
            firmware = parser.parse()
            self._append_binaries(firmware)

        return True

    def extract_all(self, ignore_guid: bool = False) -> None:
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stderr(devnull):
                self._extract()
                for guid in self._info:
                    if ignore_guid or (
                        self._info[guid]["content"] is not None
                        and (guid in self._file_guids)
                    ):
                        self.binaries.append(
                            UefiBinary(
                                content=self._info[guid]["content"],
                                name=self._info[guid]["name"],
                                guid=guid,
                                ext=self._info[guid]["ext"],
                            )
                        )



from setuptools import setup

with open("requirements.txt") as f:
    REQUIRED = f.readlines()

with open("README.md", "r") as f:
    README = f.read()

setup(
    name="fwhunt_scan",
    version="2.3.6",
    author="FwHunt team",
    author_email="fwhunt@binarly.io",
    packages=["fwhunt_scan"],
    license="GPL-3.0",
    url="https://github.com/binarly-io/fwhunt-scan",
    install_requires=REQUIRED,
    description="Tools for analyzing UEFI firmware and checking UEFI modules with FwHunt rules",
    long_description=README,
    long_description_content_type="text/markdown",
    platforms=["Platform Independent"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    include_package_data=True,
    zip_safe=False,
    package_data={
        "fwhunt_scan": [
            "py.typed",
            "uefi_analyzer.pyi",
            "uefi_extractor.pyi",
            "uefi_protocols.pyi",
            "uefi_scanner.pyi",
            "uefi_smm.pyi",
            "uefi_tables.pyi",
            "uefi_te.pyi",
            "uefi_types.pyi",
            "uefi_utils.pyi",
            "__init__.pyi",
        ]
    },
    scripts=["fwhunt_scan_analyzer.py", "fwhunt_scan_docker.py"],
)



# SPDX-License-Identifier: GPL-3.0+
#
# pylint: disable=consider-using-dict-comprehension

EFI_BOOT_SERVICES_64_BIT = {
    0x00000018: "RaiseTPL",
    0x00000020: "RestoreTPL",
    0x00000028: "AllocatePages",
    0x00000030: "FreePages",
    0x00000038: "GetMemoryMap",
    0x00000040: "AllocatePool",
    0x00000048: "FreePool",
    0x00000050: "CreateEvent",
    0x00000058: "SetTimer",
    0x00000060: "WaitForEvent",
    0x00000068: "SignalEvent",
    0x00000070: "CloseEvent",
    0x00000078: "CheckEvent",
    0x00000080: "InstallProtocolInterface",
    0x00000088: "ReinstallProtocolInterface",
    0x00000090: "UninstallProtocolInterface",
    0x00000098: "HandleProtocol",
    0x000000A0: "Reserved",
    0x000000A8: "RegisterProtocolNotify",
    0x000000B0: "LocateHandle",
    0x000000B8: "LocateDevicePath",
    0x000000C0: "InstallConfigurationTable",
    0x000000C8: "LoadImage",
    0x000000D0: "StartImage",
    0x000000D8: "Exit",
    0x000000E0: "UnloadImage",
    0x000000E8: "ExitBootServices",
    0x000000F0: "GetNextMonotonicCount",
    0x000000F8: "Stall",
    0x00000100: "SetWatchdogTimer",
    0x00000108: "ConnectController",
    0x00000110: "DisconnectController",
    0x00000118: "OpenProtocol",
    0x00000120: "CloseProtocol",
    0x00000128: "OpenProtocolInformation",
    0x00000130: "ProtocolsPerHandle",
    0x00000138: "LocateHandleBuffer",
    0x00000140: "LocateProtocol",
    0x00000148: "InstallMultipleProtocolInterfaces",
    0x00000150: "UninstallMultipleProtocolInterfaces",
    0x00000158: "CalculateCrc32",
    0x00000160: "CopyMem",
    0x00000168: "SetMem",
    0x00000170: "CreateEventEx",
}

EFI_RUNTIME_SERVICES_64_BIT = {
    0x00000018: "GetTime",
    0x00000020: "SetTime",
    0x00000028: "GetWakeupTime",
    0x00000030: "SetWakeupTime",
    0x00000038: "SetVirtualAddressMap",
    0x00000040: "ConvertPointer",
    0x00000048: "GetVariable",
    0x00000050: "GetNextVariableName",
    0x00000058: "SetVariable",
    0x00000060: "GetNextHighMonotonicCount",
    0x00000068: "ResetSystem",
    0x00000070: "UpdateCapsule",
    0x00000078: "QueryCapsuleCapabilities",
    0x00000080: "QueryVariableInfo",
}

EFI_SMM_SYSTEM_TABLE2_64_BIT = {
    0x00000028: "SmmInstallConfigurationTable",
    0x00000030: "SmmIo",
    0x00000050: "SmmAllocatePool",
    0x00000058: "SmmFreePool",
    0x00000060: "SmmAllocatePages",
    0x00000068: "SmmFreePages",
    0x00000070: "SmmStartupThisAp",
    0x00000078: "CurrentlyExecutingCpu",
    0x00000080: "NumberOfCpus",
    0x00000088: "CpuSaveStateSize",
    0x00000090: "CpuSaveState",
    0x00000098: "NumberOfTableEntries",
    0x000000A0: "SmmConfigurationTable",
    0x000000A8: "SmmInstallProtocolInterface",
    0x000000B0: "SmmUninstallProtocolInterface",
    0x000000B8: "SmmHandleProtocol",
    0x000000C0: "SmmRegisterProtocolNotify",
    0x000000C8: "SmmLocateHandle",
    0x000000D0: "SmmLocateProtocol",
    0x000000D8: "SmiManage",
    0x000000E0: "SmiHandlerRegister",
    0x000000E8: "SmiHandlerUnRegister",
}

BS_PROTOCOLS = [
    "InstallProtocolInterface",
    "ReinstallProtocolInterface",
    "UninstallProtocolInterface",
    "HandleProtocol",
    "RegisterProtocolNotify",
    "OpenProtocol",
    "CloseProtocol",
    "OpenProtocolInformation",
    "ProtocolsPerHandle",
    "LocateHandleBuffer",
    "LocateProtocol",
    "InstallMultipleProtocolInterfaces",
    "UninstallMultipleProtocolInterfaces",
]

BS_PROTOCOLS_INFO_64_BIT = {
    "InstallProtocolInterface": {"offset": 0x80, "reg": "rdx"},
    "ReinstallProtocolInterface": {"offset": 0x88, "reg": "rdx"},
    "UninstallProtocolInterface": {"offset": 0x90, "reg": "rdx"},
    "HandleProtocol": {"offset": 0x98, "reg": "rdx"},
    "RegisterProtocolNotify": {"offset": 0xA8, "reg": "rcx"},
    "OpenProtocol": {"offset": 0x118, "reg": "rdx"},
    "CloseProtocol": {"offset": 0x120, "reg": "rdx"},
    "OpenProtocolInformation": {"offset": 0x128, "reg": "rdx"},
    "ProtocolsPerHandle": {"offset": 0x130, "reg": "rdx"},
    "LocateHandleBuffer": {"offset": 0x138, "reg": "rdx"},
    "LocateProtocol": {"offset": 0x140, "reg": "rcx"},
    "InstallMultipleProtocolInterfaces": {"offset": 0x148, "reg": "rdx"},
    "UninstallMultipleProtocolInterfaces": {"offset": 0x150, "reg": "rdx"},
}

OFFSET_TO_SERVICE = dict(
    [(BS_PROTOCOLS_INFO_64_BIT[s]["offset"], s) for s in BS_PROTOCOLS_INFO_64_BIT]
)

EFI_PEI_SERVICES_32_BIT = {
    0x00000018: {"name": "InstallPpi", "arg_num": 2},
    0x0000001C: {"name": "ReInstallPpi", "arg_num": 3},
    0x00000020: {"name": "LocatePpi", "arg_num": 5},
    0x00000024: {"name": "NotifyPpi", "arg_num": 2},
    0x00000028: {"name": "GetBootMode", "arg_num": 2},
    0x0000002C: {"name": "SetBootMode", "arg_num": 2},
    0x00000030: {"name": "GetHobList", "arg_num": 2},
    0x00000034: {"name": "CreateHob", "arg_num": 4},
    0x00000038: {"name": "FfsFindNextVolume", "arg_num": 3},
    0x0000003C: {"name": "FfsFindNextFile", "arg_num": 4},
    0x00000040: {"name": "FfsFindSectionData", "arg_num": 4},
    0x00000044: {"name": "InstallPeiMemory", "arg_num": 3},
    0x00000048: {"name": "AllocatePages", "arg_num": 4},
    0x0000004C: {"name": "AllocatePool", "arg_num": 3},
    0x00000050: {"name": "CopyMem", "arg_num": 3},
    0x00000054: {"name": "SetMem", "arg_num": 3},
    0x00000058: {"name": "ReportStatusCode", "arg_num": 6},
    0x0000005C: {"name": "ResetSystem", "arg_num": 1},
    0x00000060: {"name": "CpuIo", "arg_num": 1},
    0x00000064: {"name": "PciCfg", "arg_num": 1},
}
